/*
 * eximstate
 * ---------
 *
 * (c) 2002 Claranet UK Ltd. by Ollie Cook <ollie@olliecook.net>
 *
 * $Id: eximstate.c,v 1.10 2002/11/19 19:36:55 ollie Exp $
 */ 

#if HAVE_CONFIG_H
#include "../config.h"
#endif

#if HAVE_STRING_H
#include <string.h>
#endif
#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#if HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#if HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#if HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#if HAVE_NETDB_H
#include <netdb.h>
#endif
#if HAVE_SYSLOG_H
#include <syslog.h>
#endif
#if HAVE_FCNTL_H
#include <fcntl.h>
#endif
#if HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif
#if HAVE_TIME_H
#include <time.h>
#endif
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <sysexits.h>
#include "exim_info.h"
#include "eximstate.h"
#include "logging.h"
#include "process_control.h"
#include "parse.h"
#include "net.h"

struct config config;
char conf_file[MAX_BUF];

int main (int argc, char **argv) {
	struct exim_info info;
	struct conf_item *conf_item;
	pid_t pid;
	int rc, opt;
	char buf[MAX_BUF];

	/* Set up syslog and set config file to default*/
	openlog("eximstate", LOG_PID, LOG_LOCAL3);
	strncpy (conf_file, CONF_FILE, MAX_BUF - 1);

	/* Parse command-line arguments:
	 *
	 *  -x turn on debugging; no daemonise
	 *  -c alternative configuration file
	 */
	opterr = 0;
	while ((opt = getopt(argc, argv, "c:x")) != -1) {
		switch (opt) {
			case 'x':
				debugflag = 1;
				break;
			case 'c':
				strncpy (conf_file, optarg, MAX_BUF - 1);
				break;
			case '?':
			default:
				usage();
		}
	}

	/* Read in configuration file and check it's OK before continuing */
	conf_item = parse_conf (conf_file);
	if (!conf_item) {
		error ("Error parsing configuration file: %s", strerror (errno));
		exit (EX_SOFTWARE);
	}
	if (!setup_state (&config, conf_item)) {
		error ("Error setting up state with configuration values supplied. See previous errors.");
		exit (EX_CONFIG);
	} else {
		/* Free up memory pertaining to configuration item linked list */
		destroy_conf (conf_item);	
	}

	/* Check for a previous pid file */
	rc = check_pid_live (config.pid_file, "eximstate");
	switch (rc) {
		case 1:
			error ("An eximstate daemon is already running. Exiting.");
			exit (EX_TEMPFAIL);
		case 0:	
			error ("check_pid_file returned an error. Exiting.");
			exit (EX_SOFTWARE);
	}

	/* Daemonise unless started with -x flag */
	if (debugflag == 1)
		debug ("Not daemonising. Turning on debugging.");
	else {
		/* Fork a process for the daemon */
		pid = fork ();
		if (pid < 0) {
			error ("Forking failed (%s).", strerror (errno));
			exit (EX_OSERR);
		} else if (pid > 0) {
			log ("Daemonised successfully! Parent exiting. Child pid is %d.", pid);
			exit (EX_OK); 
		} 

		/* Close stdin/stdout/stderr */
		close (0);
		close (1);
		close (2);

		/* We should now be in the child process having terminated
	 	* the parent, so release the controlling terminal and set
	 	* ourselves up to handle some signals. */

		(void) setsid ();
		signal (SIGTERM, sighandler);
		signal (SIGKILL, sighandler);
		signal (SIGINT, sighandler);
		signal (SIGHUP, sighandler);
	}

	/* Make a new pid file */
	debug ("About to drop pid file at: %s", config.pid_file);
	rc = create_pid_file (config.pid_file);
	if (!rc) {
		error ("Couldn't make new pid file."); 
		exit (EX_CANTCREAT);
	}
		
	info.lastcheck = 0;

	while (1) {	
		int s; /* Socket descriptor */

		/* Wait until it's time for the next sample */
		debug ("Sleeping until %d. (Now: %d)", (info.lastcheck + config.sample_interval), time (NULL));
		while ((info.lastcheck + config.sample_interval) > time (NULL))
			sleep (1);

		/* Zero the information structure and get new info. Time how long this step takes */
		debug ("Getting queue state information");
		memset (&info, '\0', sizeof (info));
		info.timestamp = time (NULL);
		info.timetaken = time (NULL);
		rc = get_info (&info);
		info.timetaken = time (NULL) - info.timetaken;

		if (!rc) 
			error ("An error occurred getting new status information; see previously logged errors for details.");
		else {
		  /* Open a socket to the data collector host */
			debug ("Trying to connect to data collector host '%s' on port %d", config.host, config.port);
		  s = call_socket (config.host, config.port);
		  if (s == -1) {
		  	error ("Couldn't connect to the data collector. Skipping scan. (%s)", strerror (errno));
	  		info.lastcheck = time (NULL);
	  		continue;
	  	}

		/* Make sure the connection is non-blocking */
		debug ("Setting socket to non-blocking");
		if (fcntl (s, F_SETFL, O_NONBLOCK) == -1) 
			error ("Couldn't set file descriptor %d to non-blocking: %s", s, strerror (errno));

			debug ("Checking initial banner from server");
	  	/* Check 210 response to connection */
	  	if (read_line (buf, s) == 0) {
	  		error ("Failed reading connection response from server (%s:%d). Disconnecting and skipping scan.", strerror (errno), errno);
	  		disconnect (s);
	  		continue;
	  	}
	  	if (strncmp (buf, "210", 3) != 0) {
	  		error ("Server didn't reply with 210 response code: %s. Disconnecting and skipping scan.", buf);
	  		disconnect (s);
	  		continue;
	  	}
			buf[3] = '\0';
			debug ("   ... banner was '%s'", buf);
		
		  /* Identify client to the server and make sure it returns a 210 response code*/
			debug ("Identifying self to server as '%s' and checking result", config.identity);
		  send_to_peer (s, "HELO %s\r\n", config.identity);
		  if (read_line (buf, s) == 0) {
		  	error ("Failed reading HELO response from server (%s). Disconnecting and skipping scan.", strerror (errno));
		  	disconnect (s);
		  	continue;
		  }
		  if (strncmp (buf, "221", 3) != 0) {
		  	error ("Server didn't reply with 221 response code: %s. Disconnecting and skipping scan.", buf);
		  	disconnect (s);
		  	continue;
	  	}
			buf[3] = '\0';
			debug ("   ... response from server was '%s'", buf);
		
			log ("Total %d messages (of which %d frozen; %d bounces) on %s at %d. Scan took %d seconds.",
				info.queue_total,
				info.queue_frozen,
				info.queue_bounces,
				info.hostname,
				info.timestamp,
				info.timetaken);
			debug ("Sending update to server 'UPDATE %d:%d:%d:%d' and checking result", info.timestamp, info.queue_total, info.queue_frozen, info.queue_bounces);
			send_to_peer (s, "UPDATE %d:%d:%d:%d\r\n", info.timestamp, info.queue_total, info.queue_frozen, info.queue_bounces);
			if (read_line (buf, s) == 0) {
				log ("Failed reading UPDATE response from server (%s). Disconnecting.", strerror (errno));
				disconnect (s);
				continue;
			}
			buf[3] = '\0';
			debug ("   ... response from server was '%s'", buf);
			if (strncmp (buf, "220", 3) == 0) {
				log ("Server reports data stored correctly.");
			} else if (strncmp (buf, "520", 3) == 0) {
				log ("Server reports data storage failed.");
			} else {
				log ("Server's response not understood. (%s).", buf);
			}
		}

		/* If we were connected to the data collector, quit and close the socket */
		if (s != -1) {
			debug ("Disconnecting file descriptor %d from data collector server", s);
			disconnect (s);
		}
	}

	error ("We shouldn't have got here, but we did. Exiting.");
	exit (EX_SOFTWARE);
} 

/* Function:    get_info
 * Description: Calls various other functions to populate the exim_info
 *              structure.
 * Returns:     1 on success
 *              0 on error (see logs)
 */              
int get_info (struct exim_info *info) {
	int rc;
	int status = 1;

	rc = gethostname(info->hostname, MAX_BUF);
	if (rc == -1) {
		error ("Unable to get local hostname. (%s)", strerror(errno));
		status = 0;
	}

	/* Not really interested in load or system information.
         * NOTE: may remove completely

	rc = get_load_five_min (&info->sysinfo.five_min_load);
	if (rc == 0) {
		error ("Couldn't ascertain five minute load average.");
		status = 0;
	}

	rc = get_uptime (&info->sysinfo.uptime);
	if (rc == 0) {
		error ("Couldn't ascertain system uptime.");
		status = 0;
	} */

	rc = get_queue_count (config.exim_input_directory, &info->queue_total, &info->queue_frozen, &info->queue_bounces);
	if (rc == 0) {
		error ("Couldn't ascertain number of messages on queue.");
		status = 0;
	}

	info->lastcheck = time (NULL);

	return (status);
}

/* Function:    call_socket
 * Description: Calls the specified host on the specified port
 * Returns:     A descriptor referencing the socket or -1
 *              if an error occurred, in which case errno
 *              will be set.
 */

int call_socket(char *hostname, unsigned short portnum) {
	struct sockaddr_in sa;
	struct hostent *hp;
	int s;

	if ((hp = gethostbyname (hostname)) == NULL) {
		errno = h_errno;
		return (-1); 
	}

	memset (&sa,0,sizeof (sa));
	memcpy ((char *)&sa.sin_addr, hp->h_addr, hp->h_length);
	sa.sin_family = hp->h_addrtype;
	sa.sin_port = htons ((u_short)portnum);

	if ((s = socket (hp->h_addrtype, SOCK_STREAM, 0)) < 0)
		return (-1);
	if (connect (s, (struct sockaddr *) &sa, sizeof sa) < 0) {
		close (s);
		return (-1);
	}
	return (s);
}

/* Function:    disconnect
 * Description: Disconnect from server by sending QUIT
 */

void disconnect (int fd) {
	char buf[MAX_BUF];

	if (!fd)
		return;

	send_to_peer (fd, "QUIT\r\n");

	if (read_line (buf, fd) == 0) {
		log ("disconnect(): failed reading response from server (%s)", strerror (errno));
		close (fd);
		return;
	}

	if (strncmp (buf, "211", 3) != 0) {
		buf[4] = '\0';
		log ("disconnect(): Didn't get 211 response. Received (%s). Closing socket anyway.", buf);
		close (fd);
		return;
	}

	log ("Disconnected from server successfully.");
	close (fd);
	return;	
}

/* Function:    setup_state
 * Description: Fills in the global config struct based on what's in the
 *              linked list of configuration key/value pairs.
 * Returns:     1 if config set up OK, otherwise 0
 */
int setup_state (struct config *config, struct conf_item *first_item) {
	char *required_keys[] = { "exim_input_directory", "sample_interval", "pid_file", "host", "port", "identity", NULL };
	char **current = required_keys;
	char *value;
	
	/* Check all required keys are there */
	while (current && *current != NULL) {
		if (!get_conf (*current, first_item)) {
			error ("Required configuration key '%s' not supplied.", *current);
			return 0;
		}
		current++;
	}

	/* Check that keys which require an integer, have an integer value */
	value = get_conf ("sample_interval", first_item);
	if (!is_number (value)) {
		error ("Value for key 'sample_interval' is not a number: %s", value);
		return (0);
	} else {
		config->sample_interval = atoi (value);
	}
	
	value = get_conf ("port", first_item);
	if (!is_number (value)) {
		error ("Value for key 'port' is not a number: %s", value);
		return (0);
	} else {
		config->port = atoi (value);
	}

	/* Copy strings into config struct */
	strncpy (config->exim_input_directory, get_conf ("exim_input_directory", first_item), MAX_BUF);
	strncpy (config->pid_file, get_conf ("pid_file", first_item), MAX_BUF);
	strncpy (config->host, get_conf ("host", first_item), MAX_BUF);
	strncpy (config->identity, get_conf ("identity", first_item), MAX_BUF);

	return (1);
}

/* Function:    sighandler
 * Description: Our default action upon a signal is to exit, but log
 *              which signal we caught, just in case. 
 */
void sighandler (int sig) {
	struct conf_item *conf_item;

	if (sig == SIGHUP) {
		/* Read in configuration file and check it's OK before continuing */
		conf_item = parse_conf (conf_file);
		if (!conf_item) {
			error ("Error re-parsing configuration file: %s", strerror (errno));
			exit (EX_SOFTWARE);
		}
		if (!setup_state (&config, conf_item)) {
			error ("Error setting up state with configuration values supplied. See previous errors.");
			exit (EX_CONFIG);
		} else {
			/* Free up memory pertaining to configuration item linked list */
			destroy_conf (conf_item);	
			log ("Re-read configuration correctly.");
			return;
		}
	}

	log ("Received signal (%d). Removing pidfile and exiting.", sig);
	if (remove_pid_file (config.pid_file, 1) != 1) {
		error ("Couldn't remove pid file.");
	}
	exit (EX_OK);
}

/* Function:    usage
 * Description: Describes usage
 */
void usage () {
	fprintf (stderr, "Usage: eximstate [-x] [-c <configuration file>]\n\n");
	fprintf (stderr, "-x     debugging (don't daemonise\n");
	fprintf (stderr, "-c     alternative configuration file\n");
	exit (EX_USAGE);
}
