/*
 * eximstate: Server
 * -----------------
 *
 * The eximstate server listens for connections from
 * the eximstate clients, parses the commands they
 * send, and stores the data they send in RRD databases.
 *
 * (c) 2002 Claranet UK Ltd. by Ollie Cook <ollie@olliecook.net>
 *
 * $Id: eximstated.c,v 1.8 2002/11/19 19:36:56 ollie Exp $
 */

#if HAVE_CONFIG_H
#include "../config.h"
#endif

#if HAVE_STRING_H
#include <string.h>
#endif
#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#if HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#if HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#if HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#if HAVE_FCNTL_H
#include <fcntl.h>
#endif
#if HAVE_SYSLOG_H
#include <syslog.h>
#endif
#include <stdarg.h>
#include <stdio.h>
#include <rrd.h>
#include <sysexits.h>
#include <signal.h>
#include <errno.h>
#include "logging.h"
#include "net.h"
#include "rrd_argv.h"
#include "parse.h"
#include "config.h"
#include "eximstated.h"
#include "process_control.h"

struct config config;

int main (int argc, char **argv) {
	struct sockaddr_in server_address;
	struct client *latest_client, *client, *next;
	struct conf_item *conf_item;
	int reuse = 1, rc;
	int listen_sock, highest_sock, num_read_socks;
	fd_set socks;	/* Socket set of ready sockets */
	pid_t pid;

	/* Set up syslog */
	openlog("eximstated", LOG_PID, LOG_LOCAL3);

	/* Read in configuration file and check it's OK before continuing */
	conf_item = parse_conf (CONF_FILE);
	if (!conf_item) {
		error ("Error parsing configuration file: %s", strerror (errno));
		exit (EX_SOFTWARE);
	}
	if (!setup_state (&config, conf_item)) {
		error ("Error setting up state with configuration values supplied. See previous errors.");
		exit (EX_CONFIG);
	} else {
		/* Free up memory pertaining to configuration item linked list */
		destroy_conf (conf_item);
	}

	/* Check for a previous pid file */
	rc = check_pid_live (config.pid_file, "eximstated");
	switch (rc) {
		case 1:
			error ("An eximstated daemon is already running. Exiting.");
			exit (EX_TEMPFAIL);
		case 0:
			error ("check_pid_file returned an error. Exiting.");
			exit (EX_SOFTWARE);
	}

	/* Daemonise unless started with -x flag */
	if (argc == 2 && strncmp (argv[1], "-x", 2) == 0) {
		debugflag = 1;
		debug ("Not daemonising. Turning on debugging.");
	} else {
		/* Fork a process for the daemon */
		pid = fork();
		if (pid < 0) {
			error ("Forking failed (%s).", strerror (errno));
			exit (EX_OSERR);
		} else if (pid > 0) {
			log ("Daemonised successfully! Parent exiting. Child pid is %d.", pid);
			exit (EX_OK);
		}	

		/* Close stdin/stdout/stderr */
		close (0);
		close (1);
		close (2);

		/* We should now be in the child process having terminated
		 * the parent, so release the controlling terminal and set
		 * ourselves up to handle some signals. */

		(void) setsid ();
		signal (SIGTERM, sighandler);
		signal (SIGKILL, sighandler);
		signal (SIGINT, sighandler);
		signal (SIGHUP, sighandler);
	}

	/* Make a new pid file */
	debug ("About to drop pid file at: %s", config.pid_file);
	rc = create_pid_file (config.pid_file);
	if (!rc) {
		error ("Couldn't make new pid file.");
		exit (EX_CANTCREAT);
	}

	/* Get a socket */
	if ((listen_sock = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		error ("Couldn't open listening socket: %s", strerror (errno));
		exit (EX_OSERR);
	}
	setsockopt (listen_sock, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof (reuse));

	/* Get address information and bind to socket */
	memset (&server_address, 0, sizeof (server_address));
	server_address.sin_family = AF_INET;
	server_address.sin_addr.s_addr = htonl (INADDR_ANY);
	server_address.sin_port = htons (config.port);
	if (bind (listen_sock, (struct sockaddr *) &server_address, sizeof (server_address)) == -1 ) {
		error ("Unable to bind to socket: %s", strerror (errno));
		exit (EX_OSERR);
	}

	/* Listening Queue */
	listen (listen_sock, 8);

	/* Initialise the highest socket and the latest client */
	highest_sock = listen_sock;
	latest_client = NULL;

	/* Main Loop */
	while (1) {

		/* Set up the file descriptor set */
		FD_ZERO (&socks);
		FD_SET(listen_sock, &socks);
		client = latest_client;
		while (client) {
			if (!client->ready)
				FD_SET (client->fd, &socks);
			if (client->fd > highest_sock)
				highest_sock = client->fd;
			client = client->next;
		}

		/* Block on select until there is more data to read from a socket,
		 * or there is a new connection
		 */

		if ((num_read_socks = select (highest_sock + 1, &socks, NULL, NULL, NULL)) == -1 ) {
			switch (errno) {
				default:
					error ("select() failed: %s", strerror (errno));
					exit (EX_OSERR);
				case EBADF:
					client = latest_client;
					while (client) {
						errno = 0;
						if (fcntl (client->fd, F_GETFL, 0) < 0 && errno == EBADF) {
							log ("Marking bad file descriptor for disconnection (fd: %d)", client->fd);
							client->disconnect = 1;
						}
						client = client->next;
					}
			}
		}

		/* Deal with a new connection, if required */
		if (FD_ISSET (listen_sock, &socks)) {
			client = add_new_connection (socks, highest_sock, listen_sock, &latest_client);
			if (!client) {
				error ("Couldn't accept new connection: %s (%d)", strerror (errno), errno);	
				exit (EX_OSERR);
			}
			log ("Accepted new connection from %s. (fd: %d)", inet_ntoa (client->address.sin_addr), client->fd);
#ifdef DEBUG
			display_list (latest_client);
#endif
		}

		/* Read data from each client into the client's input buffer */
		client = latest_client;
		while (client) {
			read_socket (socks, client);
			client = client->next;
		}

		/* Deal with 'ready' clients and disconnect those that are finished */
		client = latest_client;
		while (client) {
			next = client->next;	/* Because we might free client by disconnect_client() */
			if (client->ready) {
				struct command_list *cmd_list, *temp;
#ifdef DEBUG
				log ("Received data (%s) (fd: %d)", client->inbuf, client->fd);
#endif
				client->ipos = 0;
				client->ready = 0;
				cmd_list = split_to_commands (client->inbuf);
				if (cmd_list == NULL) {
					log ("Couldn't split '%s' into command list.", client->inbuf);
				} else {
					temp = cmd_list;
					while (cmd_list) {
#ifdef DEBUG
						log ("Dealing with command: %s", cmd_list->command);
#endif
						action_command (client, &latest_client, cmd_list->command);
						cmd_list = cmd_list->next;
					}
					destroy_command_list (temp);
				}
			} else if (client->disconnect) {
				log ("Disconnecting client from %s. (fd: %d)", inet_ntoa (client->address.sin_addr), client->fd);
				disconnect_client (client, &latest_client);
			} else if (time (NULL) - client->lastcommand > config.timeout) {
				send_to_peer (client->fd, " 212 Closing connection due to timeout.\r\n");
				log ("No command read from client (%s) in %d seconds. Disconnecting client (fd: %d)", inet_ntoa (client->address.sin_addr), config.timeout, client->fd);
				client->disconnect = 1;
			}
			client = next;
		}
	}
}

/* Function:    add_new_connection
 * Description: Accepts a new connection on a socket and adds the
 *              client to the linked list of client connections.
 * Returns:     Pointer to new client structure, or
 *              NULL on error, setting errno
 */
struct client *add_new_connection (fd_set socks, int highest_sock, int listen_sock, struct client **latest_client) {
	int len;
	struct client *client;

	/* Allocate memory for client structure */
	client = calloc (1, sizeof (*client));
	if (!client) {
		errno = ENOMEM;
		return (NULL);	
	}

	/* Accept the connection */
	len = sizeof (client->address);
	if ((client->fd = accept (listen_sock, (struct sockaddr *) &client->address, &len)) < 0)
		return (NULL);

	if (client->fd > highest_sock)
		highest_sock = client->fd;

  /* Make sure the connection is non-blocking */
  if (fcntl (client->fd, F_SETFL, O_NONBLOCK) == -1) {
    log ("Couldn't set file descriptor %d to non-blocking: %s", client->fd, strerror (errno));
  }

	send_to_peer (client->fd, "210 Communications channel open. Proceed.\r\n");

	/* Set the last command time to the time of connection */
	client->lastcommand = time (NULL);

	/* Add client to linked list */
	client->next = *latest_client;
	client->prev = NULL;
	if (client->next)
		client->next->prev = client;
	*latest_client = client;

	return (client);
}

/* Function:    disconnect_client
 * Description: Disconnects the given client and removes the entry
 *              from the linked list
 */
void disconnect_client (struct client *client, struct client **latest_client) {
	if (client)
		/* Close the file descriptor */
		close (client->fd);

		/* Remove from linked list */
		if (client->next)
			client->next->prev = client->prev;
		if (client->prev)
			client->prev->next = client->next;
		if (client == *latest_client)
			*latest_client = client->next;

		/* Free up memory */
		free (client);
#ifdef DEBUG
		display_list (*latest_client);
#endif
}

/* Function:    display_list
 * Description: Shows the client linked list structure
 */
void display_list (struct client *latest_client) {
	int i=0;
	struct client *client;

	client = latest_client;
	while (client) {
		log ("%d: Address %p, next %p, previous %p, fd %d", i, client, client->next, client->prev, client->fd);
		client = client->next;
		i++;
	}
}

/* Function:    read_socket
 * Description: Reads data from the file descriptor specified in client into
 *              the input buffer in client
 */
void read_socket (fd_set socks, struct client *client) {
	ssize_t recvd_bytes; /* Bytes received from recv() */
	char buf[MAX_BUF];   /* Buffer for recv() */

  /* We only want to read if both the socket has data waiting, and a full
   * command has not yet been received into the input buffer
   */
	if (FD_ISSET (client->fd, &socks) && client->ready == 0) {
		recvd_bytes = recv (client->fd, buf, sizeof (buf) - 1, 0);
		if (recvd_bytes == 0) {
			log ("No data read from file descriptor. Marking client for disconnecton. (fd: %d)", client->fd);
			client->disconnect = 1;
		} else if (recvd_bytes < 0) {
			error ("Call to recv() on file descriptor (fd: %d) failed: %s. Marking client for disconnection.", client->fd, strerror (errno));
			client->disconnect = 1;
		} else {
			/* Put the data in the client's input buffer */
			strncpy (&client->inbuf[client->ipos], buf, recvd_bytes);
			client->ipos += recvd_bytes;
			client->inbuf[client->ipos] = '\0';
			/* If the last two characters in the buffer are \r\n we have a full command
			 * so set the client to ready
			 */
			if (client->inbuf[client->ipos - 2] == '\r' && client->inbuf[client->ipos - 1] == '\n') {
				client->lastcommand = time (NULL);
				client->ready = 1;
			}
		}
	}
}

/* Function:     action_command
 * Description:  Performs the necessary operations having received a command
 */
void action_command (struct client *client, struct client **latest_client, char *command) {
  /* Command: QUIT */
	if (strncmp (command, "QUIT", 4) == 0) {
		send_to_peer (client->fd, "211 Closing connection.\r\n");					
		log ("Marking client from %s for disconnect.", inet_ntoa (client->address.sin_addr));
		client->disconnect = 1;

  /* Command: HELO */
	} else if (strncmp (command, "HELO", 4) == 0) {
		if (sscanf (command, "HELO %s\r\n", client->identity) != 1) {
			send_to_peer (client->fd, "500 Unrecognised format of HELO command. Closing connection.\r\n");
			log ("Malformed HELO command from at %s. Marking for disconnect.", inet_ntoa (client->address.sin_addr));
			client->disconnect = 1;
			return;
		}	
		send_to_peer (client->fd, "221 Hello %s. Continue\r\n", client->identity);
		client->ready = 0;

  /* Command: UPDATE */
	} else if (strncmp (command, "UPDATE", 6) == 0) {
		int rc, argc;
		char data_buf[MAX_BUF], rrd_cmdbuf[MAX_BUF], **argv;

		if (sscanf (command, "UPDATE %s\r\n", data_buf) != 1) {
			send_to_peer (client->fd, "500 Unrecognised format of UPDATE command (%s). Closing connection.\r\n", command);
			log ("Malformed UPDATE command from at %s. Marking client for disconnect.", inet_ntoa (client->address.sin_addr));
			client->disconnect = 1;
			return;
		} 

		snprintf (rrd_cmdbuf, MAX_BUF - 1, "update %s/%s.rrd %s", config.data_dir, client->identity, data_buf);
		if ((argv = string_to_argv (rrd_cmdbuf, &argc)) != NULL) {
			/* Reset optind and opterr */
			optind = 0;
			opterr = 0;
			rc = rrd_update (argc, argv);
			free_argv (argv);
			if (rc == -1) {
				log ("Update of .rrd database failed: %s", rrd_get_error ());
				send_to_peer (client->fd, "520 Database update failed.\r\n");
				rrd_clear_error ();
			} else {
				log ("RRD update called with (%s).", data_buf);
				send_to_peer (client->fd, "220 Database updated. Information correctly stored.\r\n");
			}
		} else {
			log ("Couldn't convert string to argv for calling rrd_update: %s", strerror (errno));
			send_to_peer (client->fd, "502 Internal error. See server logs.\r\n");
		}
		client->ready = 0;

  /* Unknown Command */
	} else {
		send_to_peer (client->fd, "501 Unknown command (%s). Closing connection.\r\n", command);
		log ("Unknown command received from client. Marking client from %s for disconnect.", inet_ntoa (client->address.sin_addr));
		client->disconnect = 1;
		return;
	}
}

/* Function:    split_to_commands
 * Description: Given a string, returns a linked list of command
 * Returns:     The address of the first element of a list of commands
 *              or NULL on error (setting errno)
 */

struct command_list *split_to_commands (char* command_string) {
	char *p;
	struct command_list *first_item = NULL, *last_item = NULL, *temp;
	int c = 0, len;

	if (!command_string) {
		errno = EINVAL;
		return (NULL);
	}

	p = command_string;
	len = strlen (command_string);
	while (c <= len) {
		if (strncmp (p, "\r\n", 2) == 0) {
			/* We've reached the end of a command, add to command list */
			if (first_item == NULL) {
				/* This is the first entry in the list */
				first_item = calloc (1, sizeof (*first_item));
				strncpy (first_item->command, command_string, c);
				first_item->command[c] = '\0';
				first_item->next = NULL;
				last_item = first_item;
				command_string = p + 2;
				c = 0;
			} else {
				/* Add the command to the list */
				temp = calloc (1, sizeof (*temp));
				strncpy (temp->command, command_string, c - 2);
				temp->command[c] = '\0';
				temp->next = NULL;
				last_item->next = temp;
				last_item = temp;
				command_string = p + 2;
				c = 0;
			}
		}
		c++;
		p++;
	}
	return (first_item);
}

/* Function:    destroy_command_list
 * Description: Frees up memory associated with a command_list
 *              linked list.
 */

void destroy_command_list (struct command_list *cmd_list) {
	struct command_list *temp;

	while (cmd_list != NULL) {
		temp = cmd_list;
		cmd_list = temp->next;
		free (temp);
	}
	cmd_list = NULL;
}

/* Function:    setup_state
 * Description: Fills in the global config struct based on what's in the
 *              linked list of configuration key/value pairs.
 * Returns:     1 if config set up OK, otherwise 0
 */
int setup_state (struct config *config, struct conf_item *first_item) {
  char *required_keys[] = { "pid_file", "data_dir", "port", NULL };
  char **current = required_keys;
  char *value;

  /* Check all required keys are there */
  while (current && *current != NULL) {
    if (!get_conf (*current, first_item)) {
      error ("Required configuration key '%s' not supplied.", *current);
      return 0;
    }
    current++;
  }

  /* Check that keys which require an integer, have an integer value */
  value = get_conf ("port", first_item);
  if (!is_number (value)) {
    error ("Value for key 'port' is not a number: %s", value);
    return (0);
  } else {
    config->port = atoi (value);
  }
	
	value = get_conf ("timeout", first_item);
	if (!is_number (value)) {
		error ("Value for key 'timeout' is not a number: %s", value);
		return (0);
	} else {
		config->timeout = atoi (value);
	}

  /* Copy strings into config struct */
  strncpy (config->pid_file, get_conf ("pid_file", first_item), MAX_BUF);
  strncpy (config->data_dir, get_conf ("data_dir", first_item), MAX_BUF);

  return (1);
}

/* Function:    sighandler
 * Description: Reread config on HUP or log signal and exit for any other signal.
 */
void sighandler (int sig) {
	struct conf_item *conf_item;

	if (sig == SIGHUP) {
		/* Read in configuration file and check it's OK before continuing */
		conf_item = parse_conf (CONF_FILE);
		if (!conf_item) {
			error ("Error re-parsing configuration file: %s", strerror (errno));
			exit (EX_SOFTWARE);
		}
		if (!setup_state (&config, conf_item)) {
			error ("Error setting up state with configuration values supplied. See previous errors.");
			exit (EX_CONFIG);
		} else {
			/* Free up memory pertaining to configuration item linked list */
			destroy_conf (conf_item);
			log ("Re-read configuration correctly.");
			return;
		}
	}

	log ("Received signal (%d). Removing pidfile and exiting.", sig);
	if (remove_pid_file (config.pid_file, 1) != 1) {
		error ("Couldn't remove pid file.");
	}
	exit (EX_OK);
}
