/*
 * eximstate: exitop
 * -----------------
 *
 * A console utility to list all servers and their respective queue
 * sizes.
 *
 * (c) 2002 Claranet UK Ltd. by Ollie Cook <ollie@olliecook.net>
 *
 * $Id: exitop.c,v 1.7 2002/11/19 19:36:56 ollie Exp $
 */ 

#if HAVE_CONFIG_H
#include "../config.h"
#endif

#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if HAVE_STRING_H
#include <string.h>
#endif
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#if HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#if HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#if HAVE_DIRENT_H
#include <dirent.h>
#endif
#include <ncurses.h>
#include <stdio.h>
#include <errno.h>
#include <sysexits.h>
#include <rrd.h>
#include <signal.h>
#include "config.h"
#include "rrd_argv.h"
#include "exitop.h"

int main (int argc, char **argv) {
	WINDOW *screen;
	int rows, cols, left;
	rrd_value_t total, frozen, bounces;
	char title[] = "exitop";
	char timestamp[26];
	char **identity, **identities;
	time_t stamp;

	/* Catch signals so that we can cann endwin() or we will mess up 
	 * the terminal */
	signal (SIGTERM, sighandler);
	signal (SIGKILL, sighandler);
	signal (SIGINT, sighandler);
	signal (SIGHUP, sighandler);

	/* Initialise the screen, and get its size so we can align 
	 * the text centrally */
	screen = initscr ();
	getmaxyx (screen, rows, cols);
	left = (cols - 45) / 2;

	/* Print the headers and the box around the edge of the screen */
	attron (A_BOLD);
	mvprintw (1, (cols - strlen(title))/2, "%s", title);
	attroff (A_BOLD);
	box (screen, ACS_VLINE, ACS_HLINE);
	mvprintw (3, left + 14, "total  frozen  bounces"); /* 14 = 10 (identity len) = 4 (spaces until first item) */
	refresh ();

	/* Loop forever, until a signal is caught */
	while (1) {
		int count = 0;

		/* Get a list of the identities of the hosts that we have data for */
		identities = get_identities (DATA_DIR);
		if (!identities) {
			endwin ();
			printf ("\nCouldn't find data sources in '%s': %s.\n", DATA_DIR, strerror (errno));
			exit (EX_UNAVAILABLE);
		}

		identity = identities;
		while (identity && *identity != NULL) {
			if (rows - 5 - count < 2) {
				endwin ();
				printf ("\nToo many data sources to display. Suggest increasing size of terminal if possible\n");
				exit (EX_TEMPFAIL);
			}
	 		if (get_data (*identity, &total, &frozen, &bounces)) 
				mvprintw (5 + count, left, "%10s    %5.0f   %5.0f   %5.0f", *identity, total, frozen, bounces);
			else {
				mvprintw (5 + count, left, "%10s    (failed to collect data: %s; rrd reports: %s)",
				          *identity, strerror (errno), rrd_get_error ());
				rrd_clear_error ();
			}
			identity++;
			count++;
		}
		free_identities (identities);

		/* Print timestamp, but only if there are sufficient rows */

		if (rows - 5 - count > 2) {
			stamp = time (NULL);
			strncpy (timestamp, ctime ((time_t *) &stamp), 24);
			timestamp[24] = '\0';
			mvprintw (rows - 2, (cols - 24) / 2, "%s", timestamp);
		}
		refresh ();
		sleep (60);
	}
}

int get_data (char *identity, rrd_value_t *total, rrd_value_t *frozen, rrd_value_t *bounces) {
	time_t start, end;
	unsigned long step, ds_cnt;
	rrd_value_t *data;
	char **ds_namv, **argv, command[MAX_BUF];
	int argc, i;

	snprintf (command, MAX_BUF - 1, "fetch %s/%s.rrd LAST --start %d", DATA_DIR, identity, time (NULL) - 300);
	if ((argv = string_to_argv (command, &argc)) != NULL) {
		/* Reset optind and opterr */
		optind = 0;
		opterr = 0;
		if (rrd_fetch (argc, argv, &start, &end, &step, &ds_cnt, &ds_namv, &data) != -1) { 
			/* Check the data sources are as we expect to find them */
			if (ds_cnt != 3 ||
			    strncmp (ds_namv[0], "total", 5) ||
			    strncmp (ds_namv[1], "frozen", 6) ||
			    strncmp (ds_namv[2], "bounces", 7)) {
				errno = EINVAL;
				free_argv (argv);
				return (0);
			}

			/* The first three items of 'data' are what we're interested in */	
			*total = *(data);
			*frozen = *(data+1);
			*bounces = *(data+2);

			/* Free up all the memory that rrd_fetch allocated */	
			for (i=0; i<ds_cnt; i++) 
					free (ds_namv[i]);
			free (ds_namv);
			free (data);
			free_argv (argv);
			return (1);
		}
		free_argv (argv);
	}
	return (0);
}

/* Function:    sighandler
 * Description: Signal handler. Used to shut down cleanly
 */
void sighandler (int sig) {
	endwin ();
	printf ("\nExiting on signal: %d\n", sig);
	exit (EX_OK);
}

/* Function:    get_identities
 * Description: Looks in 'directory' for all .rrd files and 
 *              returns an array of pointers to the names
 * Returns:     A pointer or NULL on error. 
 */

char **get_identities (char *directory) {
	struct dirent *entry;
	DIR *dir;
	char **identities = NULL;
	int count = 0;
	struct stat sb;
	char filename[MAX_BUF];

	if ((dir = opendir (directory)) == NULL)
		return (NULL);
		
	while ((entry = readdir (dir)) != NULL) {
		int len;

		len = strlen (entry->d_name);

		/* stat() file so we can test if it's a regular file */
		snprintf (filename, MAX_BUF, "%s/%s", directory, entry->d_name);
		if (stat (filename, &sb) < 0)
			return (NULL);

		/* We are only concerned with entries whose last four letters
		 * are .rrd and whose type is DT_REG) */
		if (entry->d_name[len - 1] == 'd' &&
		    entry->d_name[len - 2] == 'r' &&
		    entry->d_name[len - 3] == 'r' &&
		    entry->d_name[len - 4] == '.' &&
 		    (sb.st_mode & S_IFREG)) {
			count++;
			if ((identities = (char **) realloc (identities, (count + 1) * sizeof (void *))) == NULL) {
				errno = ENOMEM;
				return (NULL);
			}
			if ((identities[count - 1] = (char *) malloc ((len * sizeof (char *)) - 3 )) == NULL) {
				errno = ENOMEM;
				return (NULL);
			}
			strncpy (identities[count - 1], entry->d_name, len - 4);
			identities[count - 1][len - 4] = '\0';
		}
	}

	closedir (dir);

	if (count > 0)
		identities[count] = NULL;
	
	return (identities);
}

/* Function:    free_identities
 * Description: Frees memory allocated by get_identities
 */

void free_identities (char **identities) {
	char **identity;

	identity = identities;
	while (identity && *identity) {
		free (*identity);
		identity++;
	}
	free (identities);
}
