/*
 * eximstate: Configuration File Parsing Functions
 * -----------------------------------------------
 *
 * Configuration file parsing functions that are common to bothi
 * daemon and client
 *
 * (c) 2002 Claranet UK Ltd. by Ollie Cook <ollie@olliecook.net>
 *
 * $Id: parse.c,v 1.2 2002/11/19 19:36:56 ollie Exp $
 */

#if HAVE_CONFIG_H
#include "../config.h"
#endif

#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#if HAVE_STRING_H
#include <string.h>
#endif
#include <errno.h>
#include <signal.h>
#include <sys/signal.h>
#include <sysexits.h>
#include <stdio.h>
#include "eximstate.h"
#include "process_control.h"
#include "logging.h"

/* Function:    parse_conf
 * Description: Parses a configuration file of key/value
 *              pair in the following format into a linked
 *              list:
 *
 *                <key> <whitespace> <value>
 *
 *              Comment character is # and any line starting
 *              with one is ignored
 *
 *              The first character of the key must be in the
 *              first column, or the line will be skipped.
 *
 * Returns:     A pointer to the first item in the linked
 *              list of key/value pairs, or NULL on error
 *              in which case errno will be set.
 */

struct conf_item *parse_conf (char *configuration_file)
{
	char *key, *value, *p;
	char buf[MAX_BUF];
	FILE *fp;
	int line = 0;
	struct conf_item *first_item = NULL, *conf_item;

	fp = fopen (configuration_file, "r");
	if (!fp)
		return (NULL);

	/* Read each line of the configuration file */
	while (fgets (buf, sizeof(buf), fp)) {
		line++;

		if (buf[0] == '#' || isspace (*buf)) 
			continue;
		
		/* Read in the key */
		key = p = buf;
		while (!isspace(*p) && *p)
			p++;

		/* Terminate the key */
		*p = 0;
		p++;
	
		/* Skip whitespace up to value */
		while (isspace (*p) && *p)
			p++;
		if (*p == 0) {
			error ("No value to key '%s' on line %d of %s",  key, line, configuration_file);
			errno = EINVAL;
			return (NULL);
		}
		
		/* Assign value */
		value = p;
		
		/* Remove trailing whitespace from value */
		while (!isspace (*p) && *p)
			p++;
		*p = 0;
		
		/* Allocate new structure for configuration item and populate it */
		conf_item = (struct conf_item*) malloc (sizeof (struct conf_item));
		if (!conf_item) {
			errno = ENOMEM;
			return (NULL);
		}
		conf_item->key = (char *) malloc (strlen (key) + 1);
		conf_item->value = (char *) malloc (strlen (value) + 1);
		if (!conf_item->key || !conf_item->value) {
			free (conf_item);
			errno = ENOMEM;
			return (NULL);
		}
		strcpy (conf_item->key, key);
		strcpy (conf_item->value, value);
		
		/* Add item to linked list */
		conf_item->next = first_item;
		first_item = conf_item;
	}
	
	fclose (fp);

	return (first_item);
}

/* Function:    get_conf
 * Description: Looks through the linked list indicated by first_item
 *              for a key/value pair with the specified key. Comparison
 *              is case insensitive.
 * Returns:     The value associated with key, or NULL if 
 *              no item with that key or no key is specified.
 */

char *get_conf (char *key, struct conf_item *first_item) {
	int len;
	struct conf_item *conf_item;
	char *return_value = NULL;

	if (!key)
		return (NULL);
	len = strlen (key);
	
	conf_item = first_item;

	while (conf_item) {
		if (!strncmp (key, conf_item->key, len + 1)) {
			return_value = conf_item->value;
			break;
		}
		conf_item = conf_item->next;
	}

	return (return_value);
}

/* Function:    destroy_conf
 * Description: Frees the pointers to keys/values in each pair, and the
 *              structure itself for all items in the configuration linked
 *              list
 */

void destroy_conf (struct conf_item *first_item) {
	struct conf_item *next;

	while (first_item) {
		next = first_item->next;
		free (first_item->key);
		free (first_item->value);
		free (first_item);
		first_item = next;
	}
}

/* Function:    is_number
 * Description: Determines if a string is comprised of only digits.
 * Returns:     1 if string is a number
 *              0 if string is not a number
 */

int is_number (char *string) {
	char *p = string;

	if (!string)
		return (0);

	while (*p) {
		if (!isdigit (*p))
			return (0);
		p++;
	}

	return (1);
}

