/*
 * eximstate: Process Control
 * --------------------------
 *
 * Functions dealing with pid file creation and removal.
 *
 * (c) 2002 Claranet UK Ltd. by Ollie Cook <ollie@olliecook.net>
 *
 * $Id: process_control.c,v 1.2 2002/11/19 19:36:56 ollie Exp $
 */

#if HAVE_CONFIG_H
#include "../config.h"
#endif

#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#if HAVE_STRING_H
#include <string.h>
#endif
#include <errno.h>
#include <signal.h>
#include <sys/signal.h>
#include <sysexits.h>
#include <stdio.h>
#include "eximstate.h"
#include "process_control.h"
#include "logging.h"

/* Function:    create_pid_file
 * Description: Stores a pid file (usually in /var/run) with
 *              the filename passes as a parameter.
 * Returns:     1 on success       
 *              0 on error
 */
int create_pid_file (char *filename) {
	FILE *fp;
	pid_t pid = getpid ();

	if ((fp = fopen (filename, "w")) == NULL) {
		error ("Couldn't open pid file for writing (%s)", strerror (errno));
		return (0);
	}
	fprintf (fp, "%d\n", pid);
	fclose (fp);
	return (1);
}

/* Function:    remove_pid_file
 * Description: Removes a pid file that was placed by create_pid_file. If flag
 *	            is 1, we flag an error if the current pid isn't the same as that
 *	            in the file.
 * Returns:     0 on error
 * 		          1 on success
 * 		          2 on pid mismatch
 */

int remove_pid_file (char *filename, int flag) {
	if (flag == 1) {	
		if (!(getpid() == get_pid_from_file (filename))) {
			error ("Current pid is %d which is different to that in the pidfile, %d. Not removing pid file.",
				getpid(),
				get_pid_from_file (filename));
			return (2);
		}
	}

	if (unlink (filename) == 0)
		return (1);
	else {
		error ("Couldn't remove pid file (%s)", strerror (errno));
		return (0);
	}
}

/* Function:    get_pid_from_file
 * Description: Given a pidfile, returns the pid, or 0 on error
 */

int get_pid_from_file (char *filename) {
	FILE *fp;
	char buf[MAX_BUF];
	pid_t pid;

	if ((fp = fopen (filename, "r")) == NULL) {
		return (0);
	}

	/* Read the process id from the file */
	fgets (buf, MAX_BUF, fp);
	fclose (fp);

	pid = atoi (buf);
	return pid;
}

/* Function:    check_pid_live
 * Description: Checks whether a process called 'name' is actually
 *	            running with pid 'pid'. If it is, its safe to assume
 *	            its the same process which left the pid file.
 * Returns:     0 on error
 *	            1 for pid alive and matches name
 *	            2 for pid alive and different name
 *	            3 for no such pid
 *	            4 for no such old pid file
 */

int check_pid_live (char *filename, char *name) {
	char buf[MAX_BUF];
	char *p;
	int pos = 0;
	FILE *fp;
	pid_t pid;

	/* Check pid file actually exists */
	if ((fp = fopen (filename, "r")) == NULL) 
		return (4);

	if ((pid = get_pid_from_file (filename)) == 0) {
		error ("Couldn't get pid from pidfile (%s).", strerror (errno));
		return (0);
	}

	/* Now read the /proc/99999/status file to get the process name */
	snprintf (buf, MAX_BUF, "/proc/%d/status", pid);
	if ((fp = fopen (buf, "r")) == NULL) {
		error ("Couldn't open /proc status file for pid %d (%s). Likely stale pidfile so removing.",
			pid,
			strerror (errno));
		if (remove_pid_file (filename, 0)) 
			log ("Stale pidfile removed sucessfully.");
		else
			error ("Error removing stale pidfile.");
		return (3);
	}
	if (fgets (buf, MAX_BUF, fp) == NULL) {
		error ("fgets() failed.");
		return (0);
	}

	fclose (fp);

	/* Parse what we read from the file to get the process name */
	p = buf;
	while (*p != '\t' && *p != ' ' && *p != '\n' ) {
		pos++;
		p++;
	}
	/* Linux has a different format for status files */
	if (strncmp (buf, "Name:", 5) == 0) {
		p = buf + 6;
		pos = 0;
		strncpy (buf, p, MAX_BUF - 1);
		buf[MAX_BUF] = '\0';
		p = buf;
		while (*p != '\n') {
			pos++;
			p++;
		}
	}
	buf[pos] = '\0';	

	return strncmp (name, buf, strlen (name)) ? 2 : 1;
}
